### Copyright 1999-2015. Parallels IP Holdings GmbH. All Rights Reserved.

from optparse import OptionParser
import os
import sys
import logging as log
import plesk_subprocess as subprocess
import re
import fileinput
from textwrap import dedent

class NotInstalledError(Exception):
    pass

def config_set(pattern, value, config, prepend=None, append=None):
    found = False
    for line in fileinput.input(config, inplace=True):
        if not found and line.strip().startswith(pattern):
            found = True
            sys.stdout.write("%s\n" % value)
        else:
            sys.stdout.write(line)

    if not found:
        with open(config, 'a') as f:
            if prepend:
                f.write("%s\n" % prepend)
            f.write("%s\n" % value)
            if append:
                f.write("%s\n" % append)

def normalize_ciphers(ciphers):
    return ciphers.strip()

def normalize_protocols(protocols_string):
    protocols = []
    exclude = []
    supported_protocols = dict(zip(map(str.lower, SUPPORTED_PROTOCOLS), SUPPORTED_PROTOCOLS))

    for p in re.split(r"\s+|:", protocols_string):
        try:
            if p.startswith('!'):
                exclude.append(supported_protocols[p[1:].lower()])
            else:
                protocols.append(supported_protocols[p.lower()])
        except KeyError:
            log.warning("Ignoring unsuppored protocol %s", p)

    if protocols and exclude:
        raise ValueError("Illegal mix of enable and disable(!) protocols in %s", protocols_string)

    if protocols:
        return protocols

    if exclude:
        return [x for x in SUPPORTED_PROTOCOLS if not x in exclude]

def condrestart(service):
    """ restart service if it started, or do nothing """
    try:
        PRODUCT_ROOT_D = "/opt/psa"
        pleskrc = os.path.join(PRODUCT_ROOT_D, 'admin/sbin/pleskrc')
        subprocess.check_call([pleskrc, service, "status"])
    except subprocess.CalledProcessError, e:
        log.debug("Service %s stopped (%d), no reload", service, e.returncode)
        return

    log.debug("reload service %s", service)
    subprocess.check_call([pleskrc, service, "reload"])

def get_ciphers():
    r = re.compile(r'^ssl-ciphers:\s(.*?)\s*$', re.MULTILINE)
    output = subprocess.check_output(['/usr/local/psa/bin/server_pref', '--show'])
    m = r.search(output)
    return m.group(1)

def get_protocols():
    r = re.compile(r'^ssl-protocols:\s(.*?)\s*$', re.MULTILINE)
    output = subprocess.check_output(['/usr/local/psa/bin/server_pref', '--show'])
    m = r.search(output)
    return m.group(1)

def nginx_action(ciphers, protocols):
    """ configure nginx by writing specific config """
    if not os.path.isfile("/etc/nginx/nginx.conf"):
        raise NotInstalledError("nginx")

    template = """\
            ssl_ciphers %s;
            ssl_protocols %s;
            ssl_prefer_server_ciphers on;
    """

    with open("/etc/nginx/conf.d/ssl.conf", "w") as f:
        f.write(dedent(template % (ciphers, ' '.join(protocols))))

    condrestart("nginx")

def sw_cp_server_action(ciphers, protocols):
    """ configure sw-cp-server (nginx) by writing specific config """

    template = """\
        ssl_ciphers %s;
        ssl_protocols %s;
        ssl_prefer_server_ciphers on;
    """

    with open("/etc/sw-cp-server/conf.d/ssl.conf", "w") as f:
        f.write(dedent(template % (ciphers, ' '.join(protocols))))

    condrestart("sw-cp-server")

def courier_action(ciphers, protocols):
    """ configure courier-imap, by writing proper data in its config. Note, it doesn't support protocols list, and
        we write only minimal procol """

    courier_protos = {
        "SSLv3" : "SSL3+",
        "TLSv1" : "TLSv1",
        "TLSv1.1" : "TLSv1.1+",
        "TLSv1.2" : "TLSv1.2+"
    }

    courier_confdir = "/etc/courier-imap"
    if not os.path.isfile(os.path.join(courier_confdir, "imapd")):
        raise NotInstalledError("courier-imap")

    proto = courier_protos.get(protocols[0])
    for service in ["imapd-ssl", "pop3d-ssl"]:
        config_set("TLS_PROTOCOL", "TLS_PROTOCOL=%s" % proto, os.path.join(courier_confdir, service))
        config_set("TLS_CIPHER_LIST", "TLS_CIPHER_LIST=%s" % ciphers, os.path.join(courier_confdir, service))
        config_set("TLS_STARTTLS_PROTOCOL", "TLS_STARTTLS_PROTOCOL=%s" % proto, os.path.join(courier_confdir, service))

    for service in ["courier-imapd", "courier-imaps", "courier-pop3d", "courier-pop3s"]:
        condrestart(service)

def dovecot_action(ciphers, protocols):
    dovecot_confdir = "/etc/dovecot"
    dovecot_include_dir = "/etc/dovecot/conf.d"

    if not os.path.isfile(os.path.join(dovecot_confdir, "dovecot.conf")):
        raise NotInstalledError("dovecot")

    template = """\
        ssl_cipher_list = %s
        ssl_protocols = %s
    """

    with open(os.path.join(dovecot_include_dir, "11-plesk-security-ssl.conf"), "w") as f:
        f.write(dedent(template % (ciphers, ' '.join(protocols))))

    condrestart("dovecot")

def apache_action(ciphers, protocols):
    if False:
# Ubuntu 12 : https://bugs.launchpad.net/ubuntu/+source/apache2/+bug/1400473
        protocols_string = "all " + ' '.join(["-%s" % x for x in SUPPORTED_PROTOCOLS if not x in protocols])
    else:
        protocols_string = ' '.join(["+%s" % x for x in protocols])

    config = "/etc/apache2/mods-available/ssl.conf"

    prepend = "<IfModule mod_ssl.c>"
    append = "</IfModule>"

    config_set("SSLProtocol", "SSLProtocol %s" % protocols_string, config, prepend, append)
    config_set("SSLCipherSuite", "SSLCipherSuite %s" % ciphers, config, prepend, append)

    condrestart("apache2")

def proftpd_action(ciphers, protocols):
    """ configure proftpd by writing specific config into proftpd include dir"""
    if not os.path.isfile("/etc/proftpd.conf"):
        raise NotInstalledError("proftpd")

    template = """\
    <IfModule mod_tls.c>
        TLSCipherSuite %s
        TLSProtocol %s
    </IfModule>
    """

    with open("/etc/proftpd.d/ssl.conf", "w") as f:
        f.write(dedent(template % (ciphers, ' '.join(protocols))))

def postfix_action(ciphers, protocols):
    postconf = "/usr/sbin/postconf"

    if not os.path.isfile(postconf):
        raise NotInstalledError("postfix")

# See https://bettercrypto.org/static/applied-crypto-hardening.pdf
    subprocess.check_call([postconf, "-e", "smtpd_tls_mandatory_protocols=%s" % ' '.join(protocols)])
    subprocess.check_call([postconf, "-e", "smtpd_tls_protocols=%s" % ' '.join(protocols)])

    subprocess.check_call([postconf, "-e", "smtpd_tls_ciphers=medium"])
    subprocess.check_call([postconf, "-e", "smtpd_tls_mandatory_ciphers=medium"])
    subprocess.check_call([postconf, "-e", "tls_medium_cipherlist=%s" % ciphers])

    condrestart("postfix")

ACTIONS = {
    "nginx" : nginx_action,
    "sw-cp-server" : sw_cp_server_action,
    "courier" : courier_action,
    "dovecot" : dovecot_action,
    "apache" : apache_action,
    "proftpd" : proftpd_action,
    "postfix" : postfix_action
}

SERVICES = ['nginx', 'sw-cp-server', 'courier', 'dovecot', 'apache', 'proftpd', 'postfix']
# Note, protocols it should be sorted by strength, as we need to check least sufficient protocol version for courier
SUPPORTED_PROTOCOLS = "TLSv1 TLSv1.1 TLSv1.2".split()

def do_main():
    parser = OptionParser(usage="Usage: %prog [-v]")
    parser.add_option("-v", "--verbose", action="count", help="Increase verbosity (can be specified multiple times).")
    parser.add_option("-c", "--ciphers", default=get_ciphers(), type="string", help="Set selected SSL ciphers.")
    parser.add_option("-p", "--protocols", default=get_protocols(), type="string", help="Set selected SSL protocols.")
    parser.add_option("-s", "--services", type="string", help="Configure only selected services.")
    (options, args) = parser.parse_args()

    log_level = log.WARNING
    if options.verbose == 1:
        log_level = log.INFO
    elif options.verbose >= 2:
        log_level = log.DEBUG

    log.basicConfig(format='%(levelname)s:%(message)s', level=log_level)

    ciphers = normalize_ciphers(options.ciphers)
    protocols = normalize_protocols(options.protocols)

    services = SERVICES
    if options.services:
        services = options.services.split()

    if not protocols:
        raise ValueError("No supported protocols supplied")

    log.debug("ciphers: %s", ciphers)
    log.debug("protocols: %s", ', '.join(protocols))

    for service in services:
        try:
            f = ACTIONS.get(service)
            if f:
                f(ciphers, protocols)
                log.debug("* %s", service)
            else:
                log.warning("Ignore unknown service %s", service)
        except NotInstalledError, e:
            log.debug("- %s\t\tnot installed", e)
        except Exception, ex:
            log.error("failed to configure %s service: %s", service, ex)
            log.debug(ex, exc_info=True)

    return 0

def main():
    """ entry point """
    try:
        return do_main()
    except SystemExit:
        raise
    except Exception, ex:
        log.error('%s', ex)
        log.debug(ex, exc_info=True)
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :

